import re
import os

CLANG_FORMAT_VERSION = '8'
YAPF_VERSION = '0.24.0'
FORMATTABLE_LANGUAGES = ['c', 'py']
CHECKED_FILES = re.compile(
    r'.*(\.(i?[ch](pp|xx|c)?|cc|hh|py|C|cmake|[yx]ml|qm[ts]|dtd|xsd|ent|bat|[cz]?sh|js|html?)|'
    r'CMakeLists.txt|Jenkinsfile)$')

DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')



from lb_dev_formatting import constants
from lb_utils.file_utils import FileUtils
from lb_utils.shell_utils import ShellUtils
import logging
from pyOptional import Optional
import os
import sys

log = logging.getLogger(__name__)


class CFormatter:

    clang_format_commands = [
        'clang-format-{}', 'lcg-clang-format-{}', 'lcg-clang-format-{}.0',
        'lcg-clang-format-{}.0.0'
    ]

    def __init__(self, clang_version):
        self.clang_version = clang_version
        self.is_clang_format_available = None
        self.clang_format_command = ''
        self.check_if_format_command_is_available()
        log.debug(
            'c formatter initialized with clang-format version: {}. Is formatting command available: {}'
            .format(self.clang_version, self.is_clang_format_available))

    def check_if_format_command_is_available(self):
        possible_clang_format_commands = [
            command.format(self.clang_version)
            for command in self.clang_format_commands
        ]
        self.clang_format_command = self.find_command(
            possible_clang_format_commands)
        if not self.clang_format_command:
            self.is_clang_format_available = False
            log.debug('No clang format found for version {}'.format(
                self.clang_version))
        else:
            self.is_clang_format_available = True

    def format_file(self, file):
        if self.is_clang_format_available:
            log.debug('formatting file: {}'.format(file))
            with open(file, 'rb') as f:
                before = f.read()
                if FileUtils.is_empty(file):
                    log.info('file is empty')
                    output = Optional(b'')
                else:
                    output = self.apply_formating(
                        before,
                        file,
                    )

                if output.is_present():
                    after = output.get()

                return Optional({'before': before, 'after': after})
        else:
            log.error('c formatter is not available, skipping file')
            return Optional.empty()

    def find_command(self, names):
        from whichcraft import which
        try:
            return next(path for path in (which(name) for name in names)
                        if path)
        except StopIteration:
            return None

    def apply_formating(self, input, path, retry=True):
        '''
        Apply formatting rules to a file.

        :param input: content of the file to format
        :param path: name of the file
        :param retry: boolean flag to tell if we have to retry the formatting
                      with a slightly modified name (see
                      https://gitlab.cern.ch/lhcb-core/LbDevTools/issues/20)

        :return: modified file, exception in case of problems
        '''
        import logging
        from subprocess import CalledProcessError
        path = os.path.join(constants.DATA_DIR, 'default.clang-format')
        FileUtils.ensure_file_exists_in_repository(
            path, '.clang-format',
            open('{}/default.clang-format'.format(constants.DATA_DIR)))

        try:
            formatting_command = [
                self.clang_format_command, '-style=file',
                '-fallback-style=none', '-assume-filename=' + path
            ]
            log.debug('formatting command is: {}'.format(formatting_command))
            return Optional(ShellUtils.run_in_shell(formatting_command, input))

        except CalledProcessError:
            if path.endswith('.h') and retry:
                # this is a workaround for cases where clang-format does
                # not correctly detect the language
                try:
                    alias = path + 'h'
                    logging.info('retry formatting of %s as %s', path, alias)
                    return self.apply_formating(input, alias, False)
                except CalledProcessError:
                    # ignore failures in the retry
                    pass
            raise  # raise original exception

    def format_in_pipeline(self):
        FileUtils.ensure_file_exists_in_repository(
            os.getcwd(), '.clang-format',
            open(os.path.join(constants.DATA_DIR, 'default.clang-format')))
        formatting_command = [
            self.clang_format_command, '-style=file', '-fallback-style=none'
        ]
        log.debug('formatting command is: {}'.format(formatting_command))
        print(ShellUtils.run_in_shell(formatting_command,
                                      sys.stdin.read().encode()),
              end='')



from lb_dev_formatting.commands.format import format_files
from lb_utils.git_utils import GitUtils
from lb_utils.log_utils import set_up_logging
from lb_dev_formatting import constants

import logging

log = logging.getLogger(__name__)

def format():
    set_up_logging(10)
    files = GitUtils.get_all_files_tracked_in_repo(None)
    return format_files(files, constants.CLANG_FORMAT_VERSION, '0.24.0', True, None, False, None)


from lb_dev_formatting.formatter.python_formatter import PythonFormatter
from lb_dev_formatting.formatter.c_formatter import CFormatter
from lb_dev_formatting import constants
from lb_utils.file_utils import FileUtils

from pyOptional import Optional
import logging

log = logging.getLogger(__name__)


class FormatterFacade:
    def __init__(self,
                 yapf_version=constants.YAPF_VERSION,
                 clang_version=constants.CLANG_FORMAT_VERSION,
                 shouldReturnDiff=False,
                 shouldOverwrite=True):
        log.debug(
            'FormatterFacade initialized with yapf version: {}, clang format version {}, shouldReturnDiff {} and shouldOverwrite {}'
            .format(yapf_version, clang_version, shouldReturnDiff,
                    shouldOverwrite))
        self.python_formatter = PythonFormatter(yapf_version)
        self.c_formatter = CFormatter(clang_version)
        self.shouldOverwrite = shouldOverwrite
        self.shouldReturnDiff = shouldReturnDiff
        self.files_formatted = 0

    def format_file(self, file_path):
        log.debug('formatting file {}'.format(file_path))
        formatter = self.get_appropriate_formatter(file_path)
        if formatter is None:
            log.error('No appropriate formatter for {}'.format(file_path))
            return Optional.empty()
        else:
            result = formatter.format_file(file_path)

            return self.handle_formatting_result(result, file_path)

    def handle_formatting_result(self, result, file_path):
        if result.is_present():
            log.info('result is present')
            
            diff = result.get()
            before = diff['before']
            after = diff['after']

            need_to_apply_formatting = after != before

            if need_to_apply_formatting:
                self.files_formatted += 1

                if self.shouldReturnDiff:
                    log.info('returning result as is')
                    return result

                log.info('file does not comply with formatting rules')
                if self.shouldOverwrite:
                    log.info('overwriting file')
                    log.info('%s changed', file_path)
                    FileUtils.write_to_file(file_path, after)
                else:
                    print(file_path, 'should be changed')

        return result

    def format_in_pipeline(self, file_type):
        formatter = self.get_appropriate_formatter_by_file_extension(file_type)

        formatter.format_in_pipeline()

    def get_appropriate_formatter_by_file_extension(self, file_extension):
        log.info('choosing appropriate formatter by file extension')
        if 'c' == file_extension:
            formatter = self.c_formatter
        elif 'py' == file_extension:
            formatter = self.python_formatter
        else:
            formatter = None
            print('invalid extension {}'.format(file_extension))

        return formatter

    def get_appropriate_formatter(self, file_path):
        log.info('choosing appropriate formatter')
        if 'c' == FileUtils.get_language_family(file_path):
            formatter = self.c_formatter
        elif 'py' == FileUtils.get_language_family(file_path):
            formatter = self.python_formatter
        else:
            formatter = None
            print('invalid file: {}'.format(file_path))

        return formatter



import click
import click_plugins
import logging
import os
from subprocess import check_output, CalledProcessError
from pyOptional import Optional
from difflib import unified_diff

from lb_dev_formatting import constants
from lb_dev_formatting.exceptions.command_not_found import CommandNotFoundError
from lb_dev_formatting.exceptions.illegal_argument import IllegalArgumentError
from lb_utils.log_utils import set_up_logging
from lb_utils.file_utils import FileUtils
from lb_utils.git_utils import GitUtils
from lb_utils.shell_utils import ShellUtils
from lb_dev_formatting.formatter.formatter_facade import FormatterFacade

log = logging.getLogger(__name__)


@click.command(
    name='format',
    help='Apply formatting rules to files given (Python and C++ files)')
@click.argument('files', nargs=-1)
@click.option('--clang-format-version',
              '-c',
              default=constants.CLANG_FORMAT_VERSION,
              help='version of clang-format to use (default: {})'.format(
                  constants.CLANG_FORMAT_VERSION))
@click.option('--yapf-version',
              '-y',
              default=constants.YAPF_VERSION,
              help='version of yapf to use (default: {})'.format(
                  constants.YAPF_VERSION))
@click.option('--dry-run', '-d', is_flag=True, help='do not modify the files')
@click.option(
    '--reference',
    '-r',
    help=
    'check/format only the files select the files that have changed since the REFERENCE commit/branch'
)
@click.option('--format-patch',
              '-f',
              help='create a patch file with the changes')
@click.option('--pipe',
              '-p',
              type=click.Choice(constants.FORMATTABLE_LANGUAGES),
              help='format from stdin to stdout (allowed values: {})'.format(
                  constants.FORMATTABLE_LANGUAGES))
@click.option('-v', '--verbose', count=True)
def format_command(files, clang_format_version, yapf_version, dry_run,
                   reference, format_patch, pipe, verbose):
    set_up_logging(verbose)
    log.debug('called `format` with parameters: files={}, \
clang_format_version={}, yapf_version={}, dry_run={}, reference={}, \
format_patch={}, pipe={}'.format(files, clang_format_version, yapf_version,
                                 dry_run, reference, format_patch, pipe))
    return format_files(files, clang_format_version, yapf_version, dry_run,
                        reference, format_patch, pipe)


def format_files(files, clang_format_version, yapf_version, dry_run, reference,
                 format_patch, pipe):
    log.info('applying formatting to files')

    check_for_illegal_arguments(files, reference, format_patch, pipe)

    if not pipe and not files:
        files = FileUtils.get_files(reference)

    patch = []
    encoding_errors = []
    shouldReturnDiff = True if format_patch else False

    formatter = FormatterFacade(yapf_version, clang_format_version,
                                shouldReturnDiff, not dry_run)

    if not files:
        log.info('formatting in pipe mode')
        formatter.format_in_pipeline(pipe)
        return 0

    for file in files:
        try:
            log.debug('formatting file {}'.format(file))
            result = formatter.format_file(file)
            if result.is_present() and format_patch:
                patch = add_diff_to_patch(result.get(), file)

        except CalledProcessError as err:
            log.warning('could not format %r: %s\n%s', file, err,
                        err.output.rstrip())

        except UnicodeDecodeError as err:
            log.error('invalid encoding in %r: %s', file, err)
            encoding_errors.append(file)

        except FileNotFoundError as err:
            log.error('file "{}" is supposed to exist in this repository but was not found. Ignoring..'.format(file))
    
    if encoding_errors:
        log.info('found encoding errors')
        inform_about_encoding_errors(encoding_errors)
        return 1

    if patch:
        log.info('creating formatting patch')
        create_formatting_patch(patch, format_patch)

    if patch or encoding_errors:
        return 1

    if dry_run and formatter.files_formatted > 0:
        return 1

    return 0


def add_diff_to_patch(diff, file):
    log.info('adding formatting differences to patch')
    patch = []
    before = diff['before']
    after = diff['after']

    patch.extend(
        line if line.endswith('\n') else (line +
                                          '\n\\ No newline at end of file\n')
        for line in unified_diff(
            before.decode('utf-8').splitlines(True),
            after.decode('utf-8').splitlines(True), os.path.join('a', file),
            os.path.join('b', file)))

    return patch


def check_for_illegal_arguments(files, reference, format_patch, pipe):
    if pipe:
        if format_patch:
            raise IllegalArgumentError(
                'incompatible options -f/--format-patch and -p/--pipe')
        elif reference:
            raise IllegalArgumentError(
                'incompatible options -r/--reference and -p/--pipe')
        elif files:
            raise IllegalArgumentError(
                'cannot process explicit files in --pipe mode')

    if reference and files:
        raise IllegalArgumentError('you cannot specify files with --reference')

    return 0


def inform_about_encoding_errors(encoding_errors):
    print('=======================================',
          ' Detected files with encoding (UTF-8) errors:',
          sep='\n')
    print('', *encoding_errors, sep='\n - ')
    print('', '=======================================', sep='\n')


def create_formatting_patch(patch, format_patch):
    from email.message import Message
    msg = Message()
    msg = add_headers_to_message(msg)

    payload = get_message_payload(patch)

    try:
        payload.encode()
        charset = None
    except UnicodeEncodeError:
        charset = 'utf-8'
    msg.set_payload(payload, charset=charset)

    if format_patch == '-':
        print(msg)
    else:
        create_patch_file(format_patch, msg)

        print('=======================================',
              ' You can fix formatting with:',
              '',
              sep='\n')
        if 'CI' in os.environ:
            print('   curl {CI_PROJECT_URL}/-/jobs/{CI_JOB_ID}/'
                  'artifacts/raw/{0} | '
                  'git am'.format(format_patch, **os.environ))
        else:
            print('   git am {}'.format(format_patch))
        print('', '=======================================', sep='\n')


def add_headers_to_message(msg):
    from email.utils import formatdate
    log.info('adding headers to message for patch')
    msg.add_header('From', 'Gitlab CI <noreply@cern.ch>')
    msg.add_header('Date', formatdate())
    msg.add_header('Subject', '[PATCH] Fixed formatting')

    return msg


def get_message_payload(patch):
    log.info('adding payload to the message for patch')
    return '\n'.join([
        'patch generated by {}'.format(
            '{CI_PROJECT_URL}/-/jobs/{CI_JOB_ID}'.format(
                **os.environ) if 'CI' in os.environ else 'standalone job'), '',
        '', ''.join(patch)
    ])


def create_patch_file(format_patch, msg):
    log.info('creating patch file')
    if (os.path.dirname(format_patch)
            and not os.path.isdir(os.path.dirname(format_patch))):
        os.makedirs(os.path.dirname(format_patch))
    with open(format_patch, 'wb') as patchfile:
        patchfile.write(bytes(msg))



from lb_dev_formatting import constants
from lb_utils.shell_utils import ShellUtils
from lb_utils.file_utils import FileUtils

from pyOptional import Optional
import logging
from subprocess import check_output
import sys

log = logging.getLogger(__name__)


class PythonFormatter:
    def __init__(self, yapf_version):
        self.yapf_version = yapf_version
        self.is_yapf_available = None
        self.check_if_format_command_is_available()
        log.debug(
            'python formatter initialized with yapf version: {}. Is formatting command available: {}'
            .format(self.yapf_version, self.is_yapf_available))

    def check_if_format_command_is_available(self):
        log.info('looking for compatible yapf version in the system')
        yapf_command = ShellUtils.find_command(['yapf'])
        yapf_version = None

        if yapf_command:
            yapf_version = check_output([yapf_command,
                                     '--version']).split()[-1].decode()
        log.debug('system yapf version: {}. Self yapf version: {}'.format(
            yapf_version, self.yapf_version))
        if yapf_version != self.yapf_version:
            log.error('no yapf version {} found in the system'.format(
                self.yapf_version))
            print('Error: no yapf version {} found in the system'.format(
                self.yapf_version))
            self.is_yapf_available = False
        else:
            self.is_yapf_available = True

    def format_file(self, file):
        if self.is_yapf_available:
            log.debug('formatting file: {}'.format(file))
            with open(file, 'rb') as f:
                before = f.read()
                if FileUtils.is_empty(file):
                    log.info('file is empty')
                    output = Optional(b'')
                else:
                    output = Optional(ShellUtils.run_in_shell(['yapf'],
                                                              before))
                if output.is_present():
                    after = output.get()

                return Optional({'before': before, 'after': after})
        else:
            log.error('python formatter is not available, skipping file')
        return Optional.empty()

    def format_in_pipeline(self):
        print(ShellUtils.run_in_shell(['yapf'],
                                      sys.stdin.read().encode()),
              end='')



###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from __future__ import absolute_import
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
from sys import version_info

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='lb_dev_formatting',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.0.4',
    use_scm_version=True,

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='Formatting plugin for LHCb Development Tools',  # Required

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    # url='https://gitlab.cern.ch/lhcb-core/lb_dev',  # Optional # TBD

    # This should be your name or the name of the organization which owns the
    # project.
    author='CERN - LHCb Core Software',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='lhcb-core-soft@cern.ch',  # Optional
    license='GPLv3+',

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='!=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4',

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    # keywords='LHCb Dirac LHCbDirac',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['*.tests']),
    package_data={'': ['default.clang-format']},
    include_package_data=True,

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['click', 'click-plugins', 'pyoptional', 'whichcraft'],
    tests_require=['pytest'],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    scripts=[],  # to be filled
    entry_points='''
        [lb_dev.plugins]
        format=lb_dev_formatting.commands.format:format_command
        [console_scripts]
        format-hook=lb_dev_formatting.hooks.format:format
    ''',

    # The package can be safely distributed as a ZIP file
    zip_safe=False,
)
